/* 
 *                              D W I N E
 *                             Version 1.1
 * 
 *                   Copyright (c) Teknowledge Inc, 1986
 * 
 *                           February 23, 1986			
 *		             for M.1 Version 2 
 *
 *	FILE: DWFUNC.C
 *
 *	PURPOSE:  Provide the external code in C for the DWINE sample
 *		  knowledge system.  Interface between M.1 and
 *		  the C-to-dBASE package of subroutines, enabling M.1
 *		  to read dBASE II files.
 *
 *	NOTES: 1) These functions do not check whether records of the data
 *		  base have been deleted.   They may be modified to check
 *		  the delete flag in a record before using the data in the
 *		  record.
 *	       2) This implementation is for Computer Innovations C86, Version
 *		  2.3.
 */

#include "stdio.h"
#include "c2dbase.h"
#include "extif.h"

/*  The following buffer, Expbuf, is used for exporting values to M.1.
 *  We make it very large to handle all possible arguments.   Most likely
 *  this is bigger than needed.
 */   
#define	BUFLEN	80				/* maximum string length */
unsigned char	Expbuf[MAXARGS][BUFLEN];

/*  MCSptr is a pointer to the master control structure and current 
 *  record in dBASE II for use by the C-to-dBASE subroutines.
 */
MASTER		*MCSptr;		/* pointer 		*/
RECNO		cur_recno;		/* current record	*/
char		dbopen;			/* true if open		*/

/******************************************************************************

                             EXTINIT
 
   PURPOSE: An initialization routine that is automatically called when 
            M.1 loads external code.

******************************************************************************/

ExtInit()
{
	int	db_open();
	int	db_close();
	int	db_read();
	int	db_skip();
	int	findwine();

	/* Set up the function lookup table */
	addFunction("open",0,db_open);
	addFunction("close",1,db_close);
	addFunction("read",2,db_read);
	addFunction("skip",3,db_skip);
	addFunction("findwine",4,findwine);

	/* Allocate memory for the master pointer structure */
	if ( !dalloc(&MCSptr) ) {
		setErrornum(EXTBADMALLOC);
	}

	/* Initialize control variables */
	cur_recno = 0;
	dbopen = FALSE;
}

/******************************************************************************
 
                             EXTEXIT
 
   PURPOSE: Termination routine that is automatically called when 
            M.1 terminates.   This routines frees memory and closes
	    files.

******************************************************************************/
 
ExtExit()
{
	/* Close the database file just in case */
	if (dbopen) {
		dclose(MCSptr);
	}

	/* Free memory */
	dfree(MCSptr);
}

/******************************************************************************
 
                             DB_OPEN
 
   PURPOSE: Open the database and index files.

   INPUT from M1:
 		item 1 = database file name (without .DBF extension)	
 		item 2 = index file name (without .NDX extension)
   OUTPUT to M1:
 		item 1 = 'ok' on success, 'not found' on failure to
 			 find files. Aborts on bad import items.

******************************************************************************/

db_open()       
{
	unsigned char dbf_name[BUFLEN];		/* data base file name */
	unsigned char ndx_name[BUFLEN];		/* index file name */
	unsigned char tbuf[2*BUFLEN];		/* temp buffer */
	int errcode;

	/* Import the file names, aborting on error */
	if ( (errcode=import(1,STRING,dbf_name,BUFLEN)) != EXTOK
	    ||
	     (errcode=import(2,STRING,ndx_name,BUFLEN)) != EXTOK
	   ) {
		setErrornum(errcode);
		return;
	}
	
	/* Arguments ok. Now set up the master control structure */
	strcpy(MCSptr->m_dbf_name,dbf_name);	/* name of data base */
	strcat(MCSptr->m_dbf_name,".DBF");	/* append extension */	
	strcpy(MCSptr->m_ndx_name,ndx_name);	/* name of index */
	strcat(MCSptr->m_ndx_name,".NDX");

	MCSptr->m_ndx_here = TRUE;		/* index is present */
	MCSptr->m_mincore = CORE_MIN;		/* memory required */

	/* Open the files. return 'ok' or 'not found' depending
	   on the result.
	*/   
	if ( dopen(MCSptr) ) {
		export(1,STRING,"ok");
		dbopen = TRUE;
	} 
	else {
		derror();
		export(1,STRING,"not found");
	}
}


/******************************************************************************
 
                             DB_CLOSE
 
   PURPOSE: Close the database and index files.

   INPUT from M1:
		none.
   OUTPUT to M1:
 		item 1 = 'ok' or 'close error'.

******************************************************************************/

db_close()
{
	int rval;

	rval = dclose(MCSptr);			/* close the files */
	if ( rval ) {
		export(1,STRING,"ok");
		dbopen = FALSE;
	}
	else {
		export(1,STRING,"close error");
	}
}


/******************************************************************************
 
                             DB_READ
 
   PURPOSE: Read the current record into memory and export it to M.1.

   INPUT from M1:
 		items 1 - N: names for fields to be returned to M1 
   OUTPUT to M1:
 		items 1 - N: values of the fields as requested.
 		types are STRING if character field, FLOAT if
 		numeric field, and INT if boolean field.

******************************************************************************/

db_read()
{
	int	errcode;
	int	i;
	int	argcnt;
	int	type;
	char	tbuf[BUFLEN];

	/* Get the record.  If not end of file, export the fields to M1. 
	   Return with no results if a bad argument is found.
	*/

	if ( dget_rec(MCSptr,cur_recno )) {
		argcnt = argCount();
		for (i=0 ; i < argcnt ; i++) {
			if ((errcode=import(i+1,STRING,tbuf,BUFLEN))!=EXTOK) {
				setErrornum(errcode);
				return;
			}
			if ((type=db_getfield(tbuf,&Expbuf[i])) != NULTYPE) {
				export(i+1,type,&Expbuf[i]);
			}	
		} /* for */
	} /* if */
}

/******************************************************************************
 
                             DB_SKIP
 
   PURPOSE: Skip to the next record in key order in the data base.

   INPUT from M1:
		none.
   OUTPUT to M1:
 		item 1 - 'ok' on success else 'end of file'.

******************************************************************************/

db_skip()
{
	if ( cur_recno = get_next(MCSptr) ) {
		export(1,STRING,"ok");
	}
	else {
		export(1,STRING,"end of file");
	}
}

/******************************************************************************
 
                             FINDWINE
 
   PURPOSE:
 	Find the first entry in the DB for a wine with these characteristics.

   INPUT from M1:
 		item 1 = color
 		item 2 = body
 		item 3 = sweetness
   OUTPUT to M1:
		item 1 = 'ok' on success, 'not found' on failure.
   NOTES:
 	This function expects that the index opened has for key a concatenation
 	of the color, body, and sweetness.  Such an index can be built in 
 	dBASE II by the command:
 		index on color-body-sweetness to ...
 
 	This function updates the current record pointer.

******************************************************************************/


findwine()
{
	char color[BUFLEN];   			/* color from M1 */
	char body[BUFLEN];			/* body from M1 */
	char sweetness[BUFLEN];			/* sweetness from M1 */
	unsigned char tbuf[3*BUFLEN+1];		/* string temp  */
	int errcode;

	/* Import the parameters, aborting on error */
	if ( (errcode=import(1,STRING,color,BUFLEN)) != EXTOK
	    ||
	     (errcode=import(2,STRING,body,BUFLEN)) != EXTOK
	    ||
	     (errcode=import(3,STRING,sweetness,BUFLEN)) != EXTOK
	   ) {
		setErrornum(errcode);
		return;
	}

	/* Check if index was read ok.  If so, build the key, which is the
	   concatenatation of the properties without spaces.
	*/
	if ( !MCSptr->mndx->ndx_root ) {
		MCSptr->m_error = NDXERROR ; derror();	/* if index bad */
		export(1,STRING,"not found");		/* found' */
		return;
	}
	strcpy(tbuf,color);
	strcat(tbuf,body);
	strcat(tbuf,sweetness);
	
	/* Find the record # */
	cur_recno = find_sk(MCSptr,MCSptr->mndx->ndx_root,tbuf);
	if ( cur_recno ) {
		export(1,STRING,"ok");
	}
	else {
		export(1,STRING,"not found");
	}
}

/******************************************************************************
 
                             DB_GETFIELD
 
   PURPOSE:
	Return the value and type of a field after gathering it from the last
	record that was read.

   INPUT:
	field_name - the name of the field to be fetched
   OUTPUT:
 	field_val will have:
 		the field's string value if CHARACTER type
 		the floating point value if NUMERIC type
 		TRUE (1) or FALSE (0) if BOOLEAN type.
   RETURNS:
 	Returns the type of the argument found (NULTYPE if not found).
   NOTES:
  	This routine is invoked by db_read and not accessible from M.1.
 

******************************************************************************/

db_getfield(field_name,field_val)
char *field_name;			/* name of field to be fetched */
char *field_val; 			/* output buffer for fetched value */
{
	int	i;
	int	j;
	char	*fldstr;
	char	*cp;
	
	/* Scan until the field is found.  If so, return the
	   appropriate value.  Otherwise, return NULTYPE.
	*/   
	upper(field_name);
	for (i = 0; i < MCSptr->m_nflds; i++) {
		if ( strcmp(MCSptr->mfd[i].fld_name, field_name) ) {
			continue;
		}	       

		/* Found it */	       
	 	switch ( MCSptr->mfd[i].fld_type ) {
		case CHARACTER:			/* field is a string   */
			fldstr = &(MCSptr->mfd[i].fld_u.string);
			j = strlen(fldstr);		/* chars in field */
			cp = fldstr + j - 1;		/* end of field */
			for (; j-- && *cp == ' ' ; --cp); /* remove trailing */
			*(cp+1) = '\0'; 		/* spaces and return */
			strcpy(field_val,fldstr);	/* it in passed buf */
			return STRING;
		case NUMERIC:			/* field is a number  */
			*((float *)field_val) =		/* return as float */
			MCSptr->mfd[i].fld_u.numeric;	/* in passed addr */
			return FLOAT;
		case LOGICAL:			/* field is a boolean */
			switch ( toupper(MCSptr->mfd[i].fld_u.logical) ) {
			case 'Y':
			case 'T': 
				*((int *)field_val) = TRUE;
				break;
			case 'N':
			case 'F':
				*((int *)field_val) = FALSE;
				break;
			}
			return INT;
		}
	}
	return NULTYPE;
}


/******************************************************************************
 
                              DERROR
                              DPANIC
 
   PURPOSE:
	Intercept DPANIC and DERROR routines and use M.1 display capability
	in their place.

******************************************************************************/

#define NERRORS  32

static char *errors[NERRORS] = {
"",
"File not opened",
"Not a dBASE II file",
"Error reading field",
"Printer not opened",		
"Index file",
"File creation",
"File already exists",
"",
"",
"Illegal command line",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"Flush",
"Write",
"Node read",
"Seek",
"Record read",
"Node write",
"Node position",
"Node read",
"Close",
"Alloc",
"Seek",
""
} ;

dpanic(s,arg1)
unsigned char *s;
unsigned char arg1;
{
	char tbuf[BUFLEN];

	strcpy(tbuf,"\nC to dBASE fatal interface error: ");
	sprintf(tbuf+strlen(tbuf),s,arg1);
	m1putString(tbuf);
	exit(0x7FFF);
}

derror()
{
	char tbuf[BUFLEN];
	
	if ( MCSptr->m_error > NERRORS || *errors[MCSptr->m_error] == NULL )
		m1putString("\nUnspecified non-fatal C to dBASE error.");
	else {
		sprintf(tbuf,"\nC to dBASE non-fatal interface error: %s.\n",
	            errors[MCSptr->m_error]);
		m1putString(tbuf);
	}
}
